/* This file is auto-generated; changes will not be preserved */
#pragma once

#include "gskgpuprintprivate.h"
#include "gskgpushaderopprivate.h"
#include "gskgradientprivate.h" /* for GskRepeat */
#include "gsk/gskroundedrectprivate.h"

void
gsk_gpu_component_transfer_op (GskGpuFrame            *frame,
                               GskGpuShaderClip        clip,
                               GdkColorState          *ccs,
                               GdkColorState          *acs,
                               float                   opacity,
                               const graphene_point_t *offset,
                               GskGpuImage            *image1,
                               GskGpuSampler           sampler1,
                               const graphene_vec4_t  *params_r,
                               const graphene_vec4_t  *params_g,
                               const graphene_vec4_t  *params_b,
                               const graphene_vec4_t  *params_a,
                               const graphene_vec4_t  *table0,
                               const graphene_vec4_t  *table1,
                               const graphene_vec4_t  *table2,
                               const graphene_vec4_t  *table3,
                               const graphene_vec4_t  *table4,
                               const graphene_vec4_t  *table5,
                               const graphene_vec4_t  *table6,
                               const graphene_vec4_t  *table7,
                               const graphene_rect_t  *rect,
                               const graphene_rect_t  *tex_rect);
