/* This file is auto-generated; changes will not be preserved */
#pragma once

#include "gskgpuprintprivate.h"
#include "gskgpushaderopprivate.h"
#include "gskgradientprivate.h" /* for GskRepeat */
#include "gsk/gskroundedrectprivate.h"

void
gsk_gpu_border_op (GskGpuFrame            *frame,
                   GskGpuShaderClip        clip,
                   GdkColorState          *ccs,
                   GdkColorState          *acs,
                   float                   opacity,
                   const graphene_point_t *offset,
                   const GskRoundedRect   *outline,
                   const GdkColor         *top_border_color,
                   const GdkColor         *right_border_color,
                   const GdkColor         *bottom_border_color,
                   const GdkColor         *left_border_color,
                   const graphene_vec4_t  *border_widths,
                   const graphene_vec2_t  *inside_offset);
