/* This file is auto-generated; changes will not be preserved */
#pragma once

#include "gskgpuprintprivate.h"
#include "gskgpushaderopprivate.h"
#include "gskgradientprivate.h" /* for GskRepeat */
#include "gsk/gskroundedrectprivate.h"

void
gsk_gpu_convert_op (GskGpuFrame            *frame,
                    GskGpuShaderClip        clip,
                    GdkColorState          *ccs,
                    gboolean                ccs_premultiplied,
                    GdkColorState          *acs,
                    float                   opacity,
                    const graphene_point_t *offset,
                    GskGpuImage            *image1,
                    GskGpuSampler           sampler1,
                    gboolean                variation_opacity,
                    const graphene_rect_t  *rect,
                    const graphene_rect_t  *tex_rect);
