/* This file is auto-generated; changes will not be preserved */
#pragma once

#include "gskgpuprintprivate.h"
#include "gskgpushaderopprivate.h"
#include "gskgradientprivate.h" /* for GskRepeat */
#include "gsk/gskroundedrectprivate.h"

void
gsk_gpu_rounded_color_op (GskGpuFrame            *frame,
                          GskGpuShaderClip        clip,
                          GdkColorState          *ccs,
                          GdkColorState          *acs,
                          float                   opacity,
                          const graphene_point_t *offset,
                          const GskRoundedRect   *outline,
                          const GdkColor         *color);
