/* This file is auto-generated; changes will not be preserved */
#pragma once

#include "gskgpuprintprivate.h"
#include "gskgpushaderopprivate.h"
#include "gskgradientprivate.h" /* for GskRepeat */
#include "gsk/gskroundedrectprivate.h"

void
gsk_gpu_displacement_op (GskGpuFrame            *frame,
                         GskGpuShaderClip        clip,
                         GdkColorState          *ccs,
                         float                   opacity,
                         const graphene_point_t *offset,
                         GskGpuImage            *image1,
                         GskGpuSampler           sampler1,
                         GskGpuImage            *image2,
                         GskGpuSampler           sampler2,
                         const graphene_rect_t  *rect,
                         const graphene_rect_t  *displacement_rect,
                         const graphene_rect_t  *child_rect,
                         guint32                 x_channel,
                         guint32                 y_channel,
                         const graphene_size_t  *displace_max,
                         const graphene_size_t  *displace_scale,
                         const graphene_size_t  *displace_offset);
