/* This file is auto-generated; changes will not be preserved */
#pragma once

#include "gskgpuprintprivate.h"
#include "gskgpushaderopprivate.h"
#include "gskgradientprivate.h" /* for GskRepeat */
#include "gsk/gskroundedrectprivate.h"

void
gsk_gpu_blur_op (GskGpuFrame            *frame,
                 GskGpuShaderClip        clip,
                 GdkColorState          *ccs,
                 GdkColorState          *acs,
                 float                   opacity,
                 const graphene_point_t *offset,
                 GskGpuImage            *image1,
                 GskGpuSampler           sampler1,
                 gboolean                variation_colorize,
                 const graphene_rect_t  *rect,
                 const GdkColor         *blur_color,
                 const graphene_rect_t  *tex_rect,
                 const graphene_vec2_t  *blur_direction);
