/* This file is auto-generated; changes will not be preserved */
#pragma once

#include "gskgpuprintprivate.h"
#include "gskgpushaderopprivate.h"
#include "gskgradientprivate.h" /* for GskRepeat */
#include "gsk/gskroundedrectprivate.h"

void
gsk_gpu_cross_fade_op (GskGpuFrame            *frame,
                       GskGpuShaderClip        clip,
                       GdkColorState          *ccs,
                       float                   opacity,
                       const graphene_point_t *offset,
                       GskGpuImage            *image1,
                       GskGpuSampler           sampler1,
                       GskGpuImage            *image2,
                       GskGpuSampler           sampler2,
                       const graphene_rect_t  *rect,
                       const graphene_rect_t  *start_rect,
                       const graphene_rect_t  *end_rect,
                       float                   progress);
