/* This file is auto-generated; changes will not be preserved */
#pragma once

#include "gskgpuprintprivate.h"
#include "gskgpushaderopprivate.h"
#include "gskgradientprivate.h" /* for GskRepeat */
#include "gsk/gskroundedrectprivate.h"

void
gsk_gpu_box_shadow_op (GskGpuFrame            *frame,
                       GskGpuShaderClip        clip,
                       GdkColorState          *ccs,
                       GdkColorState          *acs,
                       float                   opacity,
                       const graphene_point_t *offset,
                       gboolean                variation_inset,
                       const graphene_rect_t  *bounds,
                       const GskRoundedRect   *outline,
                       const graphene_size_t  *shadow_offset,
                       float                   shadow_spread,
                       float                   blur_radius,
                       const GdkColor         *color);
